/* This is the widget for drawing of input points together with their 2D convex hull (CH).
 * The user can select input point distribution, number of input points and the CH algorithm. 
 * 
 * concept and design: Petr Felkel - felkepet@fel.cvut.cz
 *
 * programming: Petr Felkel   - felkepet@fel.cvut.cz
 *				Lukas Novosad - novosluk@fel.cvut.cz
 *
 * date: 21.10.2010
 *
 */

#include "drawareawidget.h"

// CGAL includes
#include <CGAL/ch_jarvis.h>
#include <CGAL/ch_eddy.h>
#include "ch_chan.h"



// constructor of the class
DrawAreaWidget::DrawAreaWidget(QWidget *parent)
	: QWidget(parent), window_center(0,0)
{
	ui.setupUi(this); 

	// initialize class variables which don't have a default constructor
	mPointGenerator = IN_DISC;
	mNumberOfPoints = 0;
	mChAlgorithm = AKL;

	// initialize the iso-rectangle and segment corners
	P1 = Point_2(-220,-120);
	P2 = Point_2( 220, 160);
}


// default destructor of the class
DrawAreaWidget::~DrawAreaWidget()
{
	
}



// slot for switching between the point generators
void DrawAreaWidget::setPointGenerator(PointGenerator pointGenerator)
{
	// update the point generator
	mPointGenerator = pointGenerator;

	// generate a new set of points
	generatePoints( mNumberOfPoints, all_points );

	// compute the convex hull according to the selected CH algorithm
	generateConvexHull( all_points, ch_points);

	// immediate redraw
	repaint();

	// create new time measurment message
	createTimeMeasurementMessage();
}


// slot for selecting the CH algorithm
void DrawAreaWidget::setChAlgorithm(ChAlgorithm chAlgorithm)
{
	// update the CH algorihm
	mChAlgorithm = chAlgorithm;

	// compute the convex hull according to the selected CH algorithm
	generateConvexHull( all_points, ch_points);

	// immediate redraw
	repaint();

    // create new time measurment message
	createTimeMeasurementMessage();
}


// slot for setting the number of points to be generated
void DrawAreaWidget::setNumberOfPoints(long numberOfPoints)
{
	// update the number of points
	mNumberOfPoints = numberOfPoints;

	// generate a new set of points
	generatePoints( mNumberOfPoints, all_points );

	// compute the convex hull according to the selected CH algorithm
	generateConvexHull( all_points, ch_points);
   
	// immediate redraw
	repaint();

	// create new time measurment message
	createTimeMeasurementMessage();
}


/*
	This method updates the time measurement message for the status bar and emits the timeMeasurementChanged signal
 */
void DrawAreaWidget::createTimeMeasurementMessage()
{
	// create an empty QString
	QString message;
	// append the measured time from the timer
	message.append(QString("%1 [s]").arg(timer.UserTime()));
	// emit the signal containing the time measurement string
	emit timeMeasurementChanged(message);
}


/*
	This method implements the resize event.
	It overrides the virtual definition from the QWidget class.
*/
void DrawAreaWidget::resizeEvent( QResizeEvent * event )
{
	window_center.setX(event->size().width()  / 2);  
	window_center.setY(event->size().height() / 2);
}

/*
	This method generates n points based on the current point distribution.
*/
void DrawAreaWidget::generatePoints( long n, std::vector<Point_2> &points )
{	
	int radius = RADIUS;
	
	// create the CGAL point generator
	CGAL::Random_points_in_disc_2< Point_2, PointCreator> genInDisc(radius);				// generated range -+radius
	CGAL::Random_points_on_circle_2< Point_2, PointCreator> genOnCircle(radius);			// generated range -+radius
	CGAL::Random_points_in_square_2< Point_2, PointCreator> genInSquare(radius);			// generated range -+radius
	CGAL::Random_points_on_square_2< Point_2, PointCreator> genOnSquare(radius);			// generated range -+radius
	CGAL::Random_points_in_iso_rectangle_2< Point_2, PointCreator> genInRectangle(P1, P2 ); // P1 = Point_2(-220,-120);
	CGAL::Random_points_on_segment_2< Point_2, PointCreator> genOnSegment(P1, P2 );			// P2 = Point_2( 220, 160);

	// erase the existing points
	points.clear();

	// allocate space for the new points
	points.reserve(n);

	// switch accordint to the current point generator
	switch( mPointGenerator )
	{
		case IN_DISC: 		CGAL::copy_n( genInDisc, n, std::back_inserter(points) );		break;
		case ON_CIRCLE:		CGAL::copy_n( genOnCircle, n, std::back_inserter(points) );		break;
		case IN_SQUARE:		CGAL::copy_n( genInSquare, n, std::back_inserter(points) );		break;
		case ON_SQUARE:		CGAL::copy_n( genOnSquare, n, std::back_inserter(points) );		break;
		case IN_RECTANGLE:	CGAL::copy_n( genInRectangle, n, std::back_inserter(points) );	break;
		case ON_SEGMENT:	CGAL::copy_n( genOnSegment, n, std::back_inserter(points) );	break;
	}

}

//-----debug
template <class ForwardIterator>
void printPoints( ForwardIterator first, ForwardIterator last, ForwardIterator prev_first, ForwardIterator prev_last )
{
	ForwardIterator fit = first;
	ForwardIterator prev_fit = prev_first;

	printf("\nConvex hull:\n--------------\n");
	while ( (fit != last) && (prev_fit != prev_last ) ) 
	{
		printf( "[ %6.1f, %6.1f ] <- [ %6.1f, %6.1f ] \n", fit->x(), fit->y(), prev_fit->x(), prev_fit->y() );
		fit++; prev_fit++;
	}
	while ( (fit != last) )
	{
		printf( "[ %6.1f, %6.1f ] <- ... \n", fit->x(), fit->y() );
		fit++; 

	}

	while ( (prev_fit != prev_last) )
	{
		printf( "........ <- [ %6.1f, %6.1f ] \n", prev_fit->x(), prev_fit->y() );
		prev_fit++;
	}
}

/*
	This method computes the convex hull based on the points passed to it.
*/
void DrawAreaWidget::generateConvexHull( std::vector<Point_2> &points, std::vector<Point_2> &ch)
{
	// reset the timer for the new measurement
	timer.Reset();

	// erase the existing convex hull points
	ch.clear();

	// switch according to the current convex hull algorithm
	switch( mChAlgorithm ) 
	{
	case AKL:
		timer.Start();
		CGAL::ch_akl_toussaint( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	case BYKAT:
		timer.Start();
		CGAL::ch_bykat( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	case EDDY:
		timer.Start();
		CGAL::ch_eddy( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	case GRAHAM_ANDREW:
		timer.Start();
		CGAL::ch_graham_andrew( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	case JARVIS:
		timer.Start();
		CGAL::ch_jarvis( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	case MY_GRAHAM:   // TODO VGE: CALL YOUR ALGORITHM HERE
		timer.Start();
		CGAL::convex_hull_points_2( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	case MY_JARVIS:   // TODO VGE: CALL YOUR ALGORITHM HERE
		timer.Start();
		CGAL::convex_hull_points_2( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	case MY_CHAN:   // TODO VGE: CALL YOUR ALGORITHM HERE
		timer.Start();
		ch_chan( points.begin(), points.end(), std::back_inserter(ch) );
		timer.Stop();
		break;
	}
	printPoints( ch.begin(), ch.end(), prev_ch.begin(), prev_ch.end() );  // DEBUG - print the computed CH

	prev_ch.erase( prev_ch.begin(), prev_ch.end() );
	std::copy( ch.begin(), ch.end(), std::back_inserter(prev_ch) );

}



/*
	This method draws all the points passed to it.
*/
void DrawAreaWidget::drawPoints( QPainter &painter, std::vector<Point_2> &points )
{
	// draw all the points in blue
	for (int i = 0; i < points.size() ; i++)
		//painter.fillRect( points[i].x() - POINT_HALFSIZE + window_center.x(), points[i].y()-POINT_HALFSIZE + window_center.y(), POINT_SIZE, POINT_SIZE, Qt::blue);
		painter.fillRect( window_center.x() + points[i].x() - POINT_HALFSIZE, window_center.y() - points[i].y()-POINT_HALFSIZE, POINT_SIZE, POINT_SIZE, Qt::blue); // reverse y
}


/*
	This method draws a polygon based on the points passed to it.
*/
void DrawAreaWidget::drawPolygon( QPainter &painter, std::vector<Point_2> &points )
{
	// remember the painter width
	int oldWidth = painter.pen().width();
	// create a green pen for the segments
	QPen pen(Qt::green);
	// set the pen width
	pen.setWidth(2);
	// set the pen for the painter
	painter.setPen(pen);
	
	// create std iterators
	std::vector<Point_2>::iterator it;
	std::vector<Point_2>::iterator it_prev;

	// iterate through all the points and draw the line segments
	it = points.begin();
	if( it != points.end() )
	{
		it_prev = it;
		it++;
		while( it != points.end() )
		{
			painter.drawLine( window_center.x() + it->x(), window_center.y() - it->y(), 
							  window_center.x() + it_prev->x(), window_center.y() - it_prev->y() );
			it_prev = it++;
		}
		painter.drawLine( window_center.x() + points.begin()->x(), window_center.y() - points.begin()->y(), 
						  window_center.x() + it_prev->x(),        window_center.y() - it_prev->y() );

	}

	// set the old pen width of the painter
	pen.setWidth(oldWidth);

	// draw the points of the polygon in red
	int i = 1;
	for( it = points.begin() ; it != points.end() ; it++ )
	{
		if ( i > 2  ) 
			painter.fillRect( window_center.x() + (*it).x() - 2*POINT_HALFSIZE, window_center.y() - (*it).y()-2*POINT_HALFSIZE, 2*POINT_SIZE, 2*POINT_SIZE, Qt::red);
		else if ( i== 2 )
			painter.fillRect( window_center.x() + (*it).x() - 2*POINT_HALFSIZE, window_center.y() - (*it).y()-2*POINT_HALFSIZE, 2*POINT_SIZE, 2*POINT_SIZE, Qt::green);
		else 
			painter.fillRect( window_center.x() + (*it).x() - 2*POINT_HALFSIZE, window_center.y() - (*it).y()-2*POINT_HALFSIZE, 2*POINT_SIZE, 2*POINT_SIZE, Qt::blue);

		i++;
	}

	// draw the points of the polygon in red
	//it = points.begin();
	//if( it != points.end() )
	//{
	//	painter.fillRect( window_center.x() + (*it).x() - 2*POINT_HALFSIZE, window_center.y() - (*it).y()-2*POINT_HALFSIZE, 2*POINT_SIZE, 2*POINT_SIZE, Qt::blue);
	//	it++;
	//	for( ; it != points.end() ; it++ )
	//	{
	//		painter.fillRect( window_center.x() + (*it).x() - 2*POINT_HALFSIZE, window_center.y() - *it).y()-2*POINT_HALFSIZE, 2*POINT_SIZE, 2*POINT_SIZE, Qt::red);
	//	}
	//}
}


/*
	This method implements the painting.
	It overrides the virtual definition from the QWidget class.
*/
void DrawAreaWidget::paintEvent(QPaintEvent * paintEvent)
{
	int radius = RADIUS;
	
	// QPainter to be used for the painting
	QPainter painter(this);
	
	// set the pen color to black
	painter.setPen(Qt::black);

	// draw the delimiter for the area of points according the the current point generator
	switch( mPointGenerator )
	{
		case IN_DISC: 	painter.drawEllipse( window_center, radius, radius ); break;
		case ON_CIRCLE:	painter.drawEllipse( window_center, radius, radius ); break;
		case IN_SQUARE:	painter.drawRect( window_center.x()-radius, window_center.y()-radius, 2*radius, 2*radius );	break;
		case ON_SQUARE:	painter.drawRect( window_center.x()-radius, window_center.y()-radius, 2*radius, 2*radius );	break;
		case IN_RECTANGLE:
			rect.setLeft(window_center.x()+P1.x());
			rect.setBottom(window_center.y()-P1.y());
			rect.setRight(window_center.x()+P2.x());
			rect.setTop(window_center.y()-P2.y());
			painter.drawRect(rect);
			break;
		case ON_SEGMENT: break;
	}

	// draw all the points
	drawPoints( painter, all_points );
	
	// draw the convex hull polygon and its points
	drawPolygon( painter, ch_points );
}